Attribute VB_Name = "modGDIPlus"
Option Explicit

Public Enum enGdiFileType
    BMP = 0
    JPG = 1
    GIF = 2
    PNG = 3
End Enum

Private Type OLEPICTDESC
   Size     As Long
   Type     As Long
   hBmp     As Long
   hPal     As Long
   Reserved As Long
End Type

Private Type GDIPlusStartupInput
    GDIPlusVersion           As Long
    DebugEventCallback       As Long
    SuppressBackgroundThread As Long
    SuppressExternalCodecs   As Long
End Type

Private Type PWMFRect16
    Left   As Integer
    Top    As Integer
    Right  As Integer
    Bottom As Integer
End Type

Private Type wmfPlaceableFileHeader
    Key         As Long
    hMf         As Integer
    BoundingBox As PWMFRect16
    Inch        As Integer
    Reserved    As Long
    CheckSum    As Integer
End Type

Public Enum GpStatus 'GDI+ Status
    OK = 0
    GenericError = 1
    InvalidParameter = 2
    OutOfMemory = 3
    ObjectBusy = 4
    InsufficientBuffer = 5
    NotImplemented = 6
    Win32Error = 7
    WrongState = 8
    Aborted = 9
    FileNotFound = 10
    ValueOverflow = 11
    AccessDenied = 12
    UnknownImageFormat = 13
    FontFamilyNotFound = 14
    FontStyleNotFound = 15
    NotTrueTypeFont = 16
    UnsupportedGDIPlusVersion = 17
    GDIPlusNotInitialized = 18
    PropertyNotFound = 19
    PropertyNotSupported = 20
    ProfileNotFound = 21
End Enum

Private Type CLSID
    Data1 As Long
    Data2 As Integer
    Data3 As Integer
    Data4(0 To 7) As Byte
End Type

Public Type ImageCodecInfo
    ClassID As CLSID
    FormatID As CLSID
    CodecName As Long           ' String Pointer; const WCHAR*
    DllName As Long             ' String Pointer; const WCHAR*
    FormatDescription As Long   ' String Pointer; const WCHAR*
    FileNameExtension As Long   ' String Pointer; const WCHAR*
    MimeType As Long            ' String Pointer; const WCHAR*
    Flags As Long
    Version As Long
    SigCount As Long
    SigSize As Long
    SigPattern As Long          ' Byte Array Pointer; BYTE*
    SigMask As Long             ' Byte Array Pointer; BYTE*
End Type

' CLSIDs (aka GUIDs) used with the EncoderParameter structure:
' NOTE:  You won 't find this enum in the GDI+ Docs. It was created as
'        part of the workaround for the constant GUID problem.
Public Enum EncoderParameterGUIDs
   EncoderCompression
   EncoderColorDepth
   EncoderScanMethod
   EncoderVersion
   EncoderRenderMethod
   EncoderQuality
   EncoderTransformation
   EncoderLuminanceTable
   EncoderChrominanceTable
   EncoderSaveFlag
End Enum

' Image encoder parameter related types
Public Enum EncoderParameterValueType
    EncoderParameterValueTypeByte = 1              ' 8-bit unsigned int
    EncoderParameterValueTypeASCII = 2             ' 8-bit byte containing one 7-bit ASCII
                                                    ' code. NULL terminated.
    EncoderParameterValueTypeShort = 3             ' 16-bit unsigned int
    EncoderParameterValueTypeLong = 4              ' 32-bit unsigned int
    EncoderParameterValueTypeRational = 5          ' Two Longs. The first Long is the
                                                    ' numerator the second Long expresses the
                                                    ' denomintor.
    EncoderParameterValueTypeLongRange = 6         ' Two longs which specify a range of
                                                    ' integer values. The first Long specifies
                                                    ' the lower end and the second one
                                                    ' specifies the higher end. All values
                                                    ' are inclusive at both ends
    EncoderParameterValueTypeUndefined = 7         ' 8-bit byte that can take any value
                                                    ' depending on field definition
    EncoderParameterValueTypeRationalRange = 8      ' Two Rationals. The first Rational
                                                    ' specifies the lower end and the second
                                                    ' specifies the higher end. All values
                                                    ' are inclusive at both ends
End Enum

' Encoder Parameter structure
Public Type EncoderParameter
   GUID As CLSID                          ' GUID of the parameter
   NumberOfValues As Long                 ' Number of the parameter values; usually 1
   Type As EncoderParameterValueType      ' Value type, like ValueTypeLONG  etc.
   Value As Long                          ' A pointer to the parameter values
End Type

' Encoder Parameters structure
Public Type EncoderParameters
   Count As Long                          ' Number of parameters in this structure; Should be 1
   Parameter As EncoderParameter          ' Parameter values; this CAN be an array!!!! (Use CopyMemory and a string or byte array as workaround)
End Type

Public Declare Function CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (dest As Any, src As Any, ByVal cb As Long) As Long
Private Declare Function lstrlenW Lib "kernel32" (ByVal psString As Any) As Long
Private Declare Function IsBadStringPtrByLong Lib "kernel32" Alias "IsBadStringPtrW" (ByVal lpsz As Long, ByVal ucchMax As Long) As Long
'Private Declare Function GlobalAlloc Lib "kernel32" (ByVal wFlags As Long, ByVal dwBytes As Long) As Long
Private Declare Function GlobalSize Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Function GlobalLock Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Function GlobalUnlock Lib "kernel32" (ByVal hMem As Long) As Long
Private Declare Function GlobalFree Lib "kernel32" (ByVal hMem As Long) As Long

' GDI Functions
Private Declare Function CreateCompatibleDC Lib "gdi32.dll" (ByVal hDC As Long) As Long
Private Declare Function OleCreatePictureIndirect Lib "olepro32.dll" (PicDesc As OLEPICTDESC, RefIID As CLSID, ByVal fPictureOwnsHandle As Long, iPic As IPicture) As Long
'Private Declare Function CreateCompatibleBitmap Lib "gdi32.dll" (ByVal hDC As Long, ByVal nWidth As Long, ByVal nHeight As Long) As Long
Private Declare Function GetDeviceCaps Lib "gdi32.dll" (ByVal hDC As Long, ByVal nIndex As Long) As Long
Private Declare Function PatBlt Lib "gdi32.dll" (ByVal hDC As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal dwRop As Long) As Long
Private Declare Function CreateBitmap Lib "gdi32.dll" (ByVal nWidth As Long, ByVal nHeight As Long, ByVal nPlanes As Long, ByVal nBitCount As Long, lpBits As Any) As Long
Private Declare Function SelectObject Lib "gdi32.dll" (ByVal hDC As Long, ByVal hObject As Long) As Long
Private Declare Function CreateSolidBrush Lib "gdi32.dll" (ByVal crColor As Long) As Long
Private Declare Function DeleteObject Lib "gdi32.dll" (ByVal hObject As Long) As Long
Private Declare Function DeleteDC Lib "gdi32.dll" (ByVal hDC As Long) As Long

' GDI+ functions
Private Declare Function GdipLoadImageFromFile Lib "GDIPlus.dll" (ByVal FileName As Long, GpImage As Long) As GpStatus
Private Declare Function GdiplusStartup Lib "gdiplus" (Token As Long, gdipInput As GDIPlusStartupInput, GDIPlusStartupOutput As Long) As Long
Private Declare Function GdipCreateFromHDC Lib "GDIPlus.dll" (ByVal hDC As Long, GpGraphics As Long) As Long
Private Declare Function GdipSetInterpolationMode Lib "GDIPlus.dll" (ByVal Graphics As Long, ByVal InterMode As Long) As Long
Private Declare Function GdipDrawImageRectI Lib "GDIPlus.dll" (ByVal Graphics As Long, ByVal img As Long, ByVal X As Long, ByVal Y As Long, ByVal Width As Long, ByVal Height As Long) As Long
Private Declare Function GdipDeleteGraphics Lib "GDIPlus.dll" (ByVal Graphics As Long) As Long
Private Declare Function GdipDisposeImage Lib "GDIPlus.dll" (ByVal Image As Long) As Long
Private Declare Function GdipCreateBitmapFromHBITMAP Lib "GDIPlus.dll" (ByVal hBmp As Long, ByVal hPal As Long, GpBitmap As Long) As Long
Private Declare Function GdipGetImageWidth Lib "GDIPlus.dll" (ByVal Image As Long, Width As Long) As Long
Private Declare Function GdipGetImageHeight Lib "GDIPlus.dll" (ByVal Image As Long, Height As Long) As Long
Private Declare Function GdipCreateMetafileFromWmf Lib "GDIPlus.dll" (ByVal hWmf As Long, ByVal deleteWmf As Long, WmfHeader As wmfPlaceableFileHeader, Metafile As Long) As Long
Private Declare Function GdipCreateMetafileFromEmf Lib "GDIPlus.dll" (ByVal hEmf As Long, ByVal deleteEmf As Long, Metafile As Long) As Long
Private Declare Function GdipCreateBitmapFromHICON Lib "GDIPlus.dll" (ByVal hIcon As Long, GpBitmap As Long) As Long
Private Declare Function GdipDrawImageRectRectI Lib "GDIPlus.dll" (ByVal Graphics As Long, ByVal GpImage As Long, ByVal dstx As Long, ByVal dsty As Long, ByVal dstwidth As Long, ByVal dstheight As Long, ByVal srcx As Long, ByVal srcy As Long, ByVal srcwidth As Long, ByVal srcheight As Long, ByVal srcUnit As Long, ByVal ImageAttributes As Long, ByVal callback As Long, ByVal callbackData As Long) As Long
Private Declare Sub GdiplusShutdown Lib "GDIPlus.dll" (ByVal Token As Long)
Private Declare Function GdipLoadImageFromStream Lib "GDIPlus.dll" (ByVal Stream As Any, ByRef Image As Long) As GpStatus
Private Declare Sub CreateStreamOnHGlobal Lib "ole32.dll" (ByRef hGlobal As Any, ByVal fDeleteOnRelease As Long, ByRef ppstm As Any)
Private Declare Function GetHGlobalFromStream Lib "ole32.dll" (ByVal pstm As Any, phGlobal As Long) As Long
Private Declare Function GdipGetImageEncodersSize Lib "GDIPlus.dll" (numEncoders As Long, Size As Long) As GpStatus
Private Declare Function GdipGetImageEncoders Lib "GDIPlus.dll" (ByVal numEncoders As Long, ByVal Size As Long, encoders As Any) As GpStatus
Private Declare Function GdipSaveImageToFile Lib "GDIPlus.dll" (ByVal Image As Long, ByVal FileName As Long, clsidEncoder As CLSID, encoderParams As Any) As GpStatus
Private Declare Function GdipSaveImageToStream Lib "GDIPlus.dll" (ByVal Image As Long, ByVal Stream As Any, clsidEncoder As CLSID, encoderParams As Any) As GpStatus

' GDI and GDI+ constants
Private Const PLANES = 14            '  Number of planes
Private Const BITSPIXEL = 12         '  Number of bits per pixel
Private Const PATCOPY = &HF00021     ' (DWORD) dest = pattern
Private Const PICTYPE_BITMAP = 1     ' Bitmap type
Private Const InterpolationModeHighQualityBicubic = 7
Private Const GDIP_WMF_PLACEABLEKEY = &H9AC6CDD7
Private Const UnitPixel = 2

Public GdiToken As Long

' Initialises GDI Plus
Public Function InitGDIPlus() As Long
    Dim gdipInit As GDIPlusStartupInput

    If GdiToken = 0 Then
        gdipInit.GDIPlusVersion = 1
        GdiplusStartup GdiToken, gdipInit, 0
    End If
    InitGDIPlus = GdiToken
End Function

' Frees GDI Plus
Public Sub FreeGDIPlus()
    If GdiToken Then
        GdiplusShutdown GdiToken
        GdiToken = 0
    End If
End Sub

' Loads the picture (optionally resized)
Public Function LoadPictureFromFile(PicFile As String, _
                                    Optional Width As Long = -1, _
                                    Optional Height As Long = -1, _
                                    Optional ByVal BackColor As Long = vbWhite, _
                                    Optional RetainRatio As Boolean = False, _
                                    Optional ByRef ImageWidth As Long, _
                                    Optional ByRef ImageHeight As Long) As IPicture

    Dim hDC     As Long
    Dim hBitmap As Long
    Dim img     As Long
    Dim bCloseGdiToken As Boolean

    If GdiToken = 0 Then InitGDIPlus: bCloseGdiToken = True

    ' Load the Image
    If GdipLoadImageFromFile(StrPtr(PicFile), img) = GpStatus.OK Then
    
        ' Calculate picture's width and height if not specified
        If Not IsMissing(ImageWidth) Or Width < 0 Or RetainRatio Then GdipGetImageWidth img, ImageWidth
        If Width < 0 Then Width = ImageWidth
        If Not IsMissing(ImageHeight) Or Height < 0 Or RetainRatio Then GdipGetImageHeight img, ImageHeight
        If Height < 0 Then Height = ImageHeight

        ' Initialise the hDC
        InitDC hDC, hBitmap, BackColor, Width, Height

        ' Resize the picture
        GdipResize img, hDC, Width, Height, RetainRatio
        GdipDisposeImage img

        ' Get the bitmap back
        GetBitmap hDC, hBitmap

        ' Create the picture
        Set LoadPictureFromFile = CreatePicture(hBitmap)
    Else
        Set LoadPictureFromFile = Nothing
    End If

Exit_:
    If bCloseGdiToken Then FreeGDIPlus: bCloseGdiToken = False
End Function

' Initialises the hDC to draw
Private Sub InitDC(hDC As Long, hBitmap As Long, BackColor As Long, Width As Long, Height As Long)
    Dim hBrush As Long

    ' Create a memory DC and select a bitmap into it, fill it in with the backcolor
    hDC = CreateCompatibleDC(ByVal 0&)
    hBitmap = CreateBitmap(Width, Height, GetDeviceCaps(hDC, PLANES), GetDeviceCaps(hDC, BITSPIXEL), ByVal 0&)
    hBitmap = SelectObject(hDC, hBitmap)
    hBrush = CreateSolidBrush(BackColor)
    hBrush = SelectObject(hDC, hBrush)
    PatBlt hDC, 0, 0, Width, Height, PATCOPY
    DeleteObject SelectObject(hDC, hBrush)
End Sub

' Resize the picture using GDI plus
Private Sub GdipResize(img As Long, hDC As Long, Width As Long, Height As Long, Optional RetainRatio As Boolean = False)
    Dim Graphics   As Long      ' Graphics Object Pointer
    Dim OrWidth    As Long      ' Original Image Width
    Dim OrHeight   As Long      ' Original Image Height
    Dim OrRatio    As Double    ' Original Image Ratio
    Dim DesRatio   As Double    ' Destination rect Ratio
    Dim DestX      As Long      ' Destination Image X
    Dim DestY      As Long      ' Destination Image Y
    Dim DestWidth  As Long      ' Destination Image Width
    Dim DestHeight As Long      ' Destination Image Height

    GdipCreateFromHDC hDC, Graphics
    GdipSetInterpolationMode Graphics, InterpolationModeHighQualityBicubic

    If RetainRatio Then
        GdipGetImageWidth img, OrWidth
        GdipGetImageHeight img, OrHeight

        OrRatio = OrWidth / OrHeight
        DesRatio = Width / Height

        ' Calculate destination coordinates
        DestWidth = IIf(DesRatio < OrRatio, Width, Height * OrRatio)
        DestHeight = IIf(DesRatio < OrRatio, Width / OrRatio, Height)
        DestX = (Width - DestWidth) / 2
        DestY = (Height - DestHeight) / 2

        GdipDrawImageRectRectI Graphics, img, DestX, DestY, DestWidth, DestHeight, 0, 0, OrWidth, OrHeight, UnitPixel, 0, 0, 0
    Else
        GdipDrawImageRectI Graphics, img, 0, 0, Width, Height
    End If
    GdipDeleteGraphics Graphics
End Sub

' Replaces the old bitmap of the hDC, Returns the bitmap and Deletes the hDC
Private Sub GetBitmap(hDC As Long, hBitmap As Long)
    hBitmap = SelectObject(hDC, hBitmap)
    DeleteDC hDC
End Sub

' Creates a Picture Object from a Handle to a bitmap
Private Function CreatePicture(hBitmap As Long) As IPicture
    Dim IID_IDispatch As CLSID
    Dim Pic           As OLEPICTDESC
    Dim iPic          As IPicture

    ' Fill in OLE IDispatch Interface ID
    IID_IDispatch.Data1 = &H20400
    IID_IDispatch.Data4(0) = &HC0
    IID_IDispatch.Data4(7) = &H46

    ' Fill Pic with necessary parts
    Pic.Size = Len(Pic)        ' Length of structure
    Pic.Type = PICTYPE_BITMAP  ' Type of Picture (bitmap)
    Pic.hBmp = hBitmap         ' Handle to bitmap

    ' Create the picture
    OleCreatePictureIndirect Pic, IID_IDispatch, True, iPic
    Set CreatePicture = iPic
End Function

' Returns a resized version of the picture
Public Function ResizePicture(p As IPicture, Width As Long, Height As Long, _
            Optional BackColor As VBRUN.SystemColorConstants = vbButtonFace, _
            Optional RetainRatio As Boolean = False) As IPicture

    Dim img       As Long
    Dim hDC       As Long
    Dim hBitmap   As Long
    Dim WmfHeader As wmfPlaceableFileHeader
    Dim bCloseGdiToken As Boolean

    If p Is Nothing Then Exit Function
    If GdiToken = 0 Then InitGDIPlus: bCloseGdiToken = True
    ' Determine pictyre type
    Select Case p.Type
        Case vbPicTypeBitmap
             GdipCreateBitmapFromHBITMAP p.Handle, ByVal 0&, img
        Case vbPicTypeMetafile
             FillInWmfHeader WmfHeader, Width, Height
             GdipCreateMetafileFromWmf p.Handle, False, WmfHeader, img
        Case vbPicTypeEMetafile
             GdipCreateMetafileFromEmf p.Handle, False, img
        Case vbPicTypeIcon
             ' Does not return a valid Image object
             GdipCreateBitmapFromHICON p.Handle, img
    End Select

    ' Continue with resizing only if we have a valid Image object
    If img Then
        InitDC hDC, hBitmap, BackColor, Width, Height
        GdipResize img, hDC, Width, Height, RetainRatio
        GdipDisposeImage img
        GetBitmap hDC, hBitmap
        Set ResizePicture = CreatePicture(hBitmap)
    End If

    If bCloseGdiToken Then FreeGDIPlus: bCloseGdiToken = False
End Function

' Fills in the wmfPlacable header
Private Sub FillInWmfHeader(WmfHeader As wmfPlaceableFileHeader, Width As Long, Height As Long)
    With WmfHeader
        .BoundingBox.Right = Width
        .BoundingBox.Bottom = Height
        .Inch = 1440
        .Key = GDIP_WMF_PLACEABLEKEY
    End With
End Sub

' Loads the picture (optionally resized)
Public Function LoadPictureFromBytes(Bytes() As Byte, _
                                        Optional Width As Long = -1, _
                                        Optional Height As Long = -1, _
                                        Optional ByVal BackColor As Long = vbWhite, _
                                        Optional RetainRatio As Boolean = False, _
                                        Optional ByRef ImageWidth As Long, _
                                        Optional ByRef ImageHeight As Long) As IPicture

    Dim Stream  As IUnknown
    Dim hDC     As Long
    Dim hBitmap As Long
    Dim img     As Long
    Dim bCloseGdiToken As Boolean

    If GdiToken = 0 Then InitGDIPlus: bCloseGdiToken = True

    ' Load the Image
    CreateStreamOnHGlobal Bytes(0), False, Stream
    If Not (Stream Is Nothing) Then
        If GdipLoadImageFromStream(Stream, img) = OK Then
            ' Calculate picture's width and height if not specified
            If Not IsMissing(ImageWidth) Or Width < 0 Then GdipGetImageWidth img, ImageWidth
            If Width < 0 Then Width = ImageWidth
            If Not IsMissing(ImageHeight) Or Height < 0 Then GdipGetImageHeight img, ImageHeight
            If Height < 0 Then Height = ImageHeight

            ' Initialise the hDC
            InitDC hDC, hBitmap, BackColor, Width, Height

            ' Resize the picture
            GdipResize img, hDC, Width, Height, RetainRatio
            GdipDisposeImage img

            ' Get the bitmap back
            GetBitmap hDC, hBitmap

            ' Create the picture
            Set LoadPictureFromBytes = CreatePicture(hBitmap)
        Else
            Set LoadPictureFromBytes = Nothing
        End If
        Set Stream = Nothing
    End If

    If bCloseGdiToken Then FreeGDIPlus: bCloseGdiToken = False
End Function

Private Function GetEncoderClsid(MimeType As String, ClassID As CLSID)
    Dim num As Long, Size As Long, i As Long
    Dim ICI() As ImageCodecInfo
    Dim Buffer() As Byte
    
    MimeType = "image/" & MimeType
    GetEncoderClsid = -1 'Failure flag
    
    ' Get the encoder array size
    GdipGetImageEncodersSize num, Size
    If Size = 0 Then Exit Function ' Failed!
    
    ' Allocate room for the arrays dynamically
    ReDim ICI(1 To num) As ImageCodecInfo
    ReDim Buffer(1 To Size) As Byte
    
    ' Get the array and string data
    GdipGetImageEncoders num, Size, Buffer(1)
    ' Copy the class headers
    CopyMemory ICI(1), Buffer(1), (Len(ICI(1)) * num)
    
    ' Loop through all the codecs
    For i = 1 To num
        ' Must convert the pointer into a usable string
        If StrComp(PtrToStrW(ICI(i).MimeType), MimeType, vbTextCompare) = 0 Then
            ClassID = ICI(i).ClassID   ' Save the class id
            GetEncoderClsid = i        ' return the index number for success
            Exit For
        End If
    Next i
    ' Free the memory
    Erase ICI
    Erase Buffer
End Function

' From www.mvps.org/vbnet...i think
'   Dereferences an ANSI or Unicode string pointer
'   and returns a normal VB BSTR
Public Function PtrToStrW(ByVal lpsz As Long) As String
    Dim sOut As String, lLen As Long

    lLen = lstrlenW(lpsz)
    If (lLen > 0) Then
        sOut = StrConv(String$(lLen, vbNullChar), vbUnicode)
        CopyMemory ByVal sOut, ByVal lpsz, lLen * 2
        PtrToStrW = StrConv(sOut, vbFromUnicode)
    End If
End Function

' This function was created as a workaround to the problem of not being
' able to declare the GUID constants for the EncoderParameter GUIDs.
Private Sub GetParameterGUID(ByVal GUIDType As EncoderParameterGUIDs, GUID As CLSID)
    ' CLSIDs (aka GUIDs) used with the EncoderParameter structure:
    With GUID
        Select Case GUIDType
           Case EncoderChrominanceTable
              .Data1 = &HF2E455DC
              .Data2 = &H9B3
              .Data3 = &H4316
              .Data4(0) = &H82
              .Data4(1) = &H60
              .Data4(2) = &H67
              .Data4(3) = &H6A
              .Data4(4) = &HDA
              .Data4(5) = &H32
              .Data4(6) = &H48
              .Data4(7) = &H1C
           Case EncoderColorDepth
              .Data1 = &H66087055
              .Data2 = &HAD66
              .Data3 = &H4C7C
              .Data4(0) = &H9A
              .Data4(1) = &H18
              .Data4(2) = &H38
              .Data4(3) = &HA2
              .Data4(4) = &H31
              .Data4(5) = &HB
              .Data4(6) = &H83
              .Data4(7) = &H37
           Case EncoderCompression
              .Data1 = &HE09D739D
              .Data2 = &HCCD4
              .Data3 = &H44EE
              .Data4(0) = &H8E
              .Data4(1) = &HBA
              .Data4(2) = &H3F
              .Data4(3) = &HBF
              .Data4(4) = &H8B
              .Data4(5) = &HE4
              .Data4(6) = &HFC
              .Data4(7) = &H58
           Case EncoderLuminanceTable
              .Data1 = &HEDB33BCE
              .Data2 = &H266
              .Data3 = &H4A77
              .Data4(0) = &HB9
              .Data4(1) = &H4
              .Data4(2) = &H27
              .Data4(3) = &H21
              .Data4(4) = &H60
              .Data4(5) = &H99
              .Data4(6) = &HE7
              .Data4(7) = &H17
           Case EncoderQuality
              .Data1 = &H1D5BE4B5
              .Data2 = &HFA4A
              .Data3 = &H452D
              .Data4(0) = &H9C
              .Data4(1) = &HDD
              .Data4(2) = &H5D
              .Data4(3) = &HB3
              .Data4(4) = &H51
              .Data4(5) = &H5
              .Data4(6) = &HE7
              .Data4(7) = &HEB
           Case EncoderRenderMethod
              .Data1 = &H6D42C53A
              .Data2 = &H229A
              .Data3 = &H4825
              .Data4(0) = &H8B
              .Data4(1) = &HB7
              .Data4(2) = &H5C
              .Data4(3) = &H99
              .Data4(4) = &HE2
              .Data4(5) = &HB9
              .Data4(6) = &HA8
              .Data4(7) = &HB8
           Case EncoderSaveFlag
              .Data1 = &H292266FC
              .Data2 = &HAC40
              .Data3 = &H47BF
              .Data4(0) = &H8C
              .Data4(1) = &HFC
              .Data4(2) = &HA8
              .Data4(3) = &H5B
              .Data4(4) = &H89
              .Data4(5) = &HA6
              .Data4(6) = &H55
              .Data4(7) = &HDE
           Case EncoderScanMethod
              .Data1 = &H3A4E2661
              .Data2 = &H3109
              .Data3 = &H4E56
              .Data4(0) = &H85
              .Data4(1) = &H36
              .Data4(2) = &H42
              .Data4(3) = &HC1
              .Data4(4) = &H56
              .Data4(5) = &HE7
              .Data4(6) = &HDC
              .Data4(7) = &HFA
           Case EncoderTransformation
              .Data1 = &H8D0EB2D1
              .Data2 = &HA58E
              .Data3 = &H4EA8
              .Data4(0) = &HAA
              .Data4(1) = &H14
              .Data4(2) = &H10
              .Data4(3) = &H80
              .Data4(4) = &H74
              .Data4(5) = &HB7
              .Data4(6) = &HB6
              .Data4(7) = &HF9
           Case EncoderVersion
              .Data1 = &H24D18C76
              .Data2 = &H814A
              .Data3 = &H41A4
              .Data4(0) = &HBF
              .Data4(1) = &H53
              .Data4(2) = &H1C
              .Data4(3) = &H21
              .Data4(4) = &H9C
              .Data4(5) = &HCC
              .Data4(6) = &HF7
              .Data4(7) = &H97
        End Select
    End With
End Sub

Public Function SavePictureBytesToFile(Bytes() As Byte, sFile As String, Optional FileType As enGdiFileType = JPG, Optional Quality As Long = 90) As Boolean
    Dim bCloseGdiToken As Boolean
    Dim img As Long, Stream As IUnknown, s As String
    Dim encoderCLSID As CLSID, encoderParams As EncoderParameters

    If Len(sFile) = 0 Then Exit Function
    If GdiToken = 0 Then InitGDIPlus: bCloseGdiToken = True

    ' Load the Image
    CreateStreamOnHGlobal Bytes(0), 0, Stream
    If Not (Stream Is Nothing) Then
        If GdipLoadImageFromStream(Stream, img) = OK Then
            ' Get the CLSID of the JPG encoder
            GetEncoderClsid GetFileType(FileType), encoderCLSID
            If FileType = enGdiFileType.JPG Then
                ' Save as a JPEG/GIF file. This format requires encoder parameters.
                ' Setup the encoder paramters
                With encoderParams
                    With .Parameter
                        .NumberOfValues = 1     ' Should be one
                        .Type = EncoderParameterValueTypeLong
                        ' Set the GUI to EncoderQuality
                        GetParameterGUID EncoderQuality, .GUID
                        .Value = VarPtr(Quality)   ' Remember: The value expects only pointers!
                    End With
                    .Count = 1     ' Only one element in this Parameter array
                End With
            End If
            SavePictureBytesToFile = (GdipSaveImageToFile(img, StrPtr(sFile), encoderCLSID, encoderParams) = GpStatus.OK)
            GdipDisposeImage img
        End If
        Set Stream = Nothing
    End If

    If bCloseGdiToken Then FreeGDIPlus: bCloseGdiToken = False
End Function

Public Function SavePictureToFile(p As IPicture, sFile As String, Optional FileType As enGdiFileType = JPG, Optional Quality As Long = 90) As Boolean
    Dim bCloseGdiToken As Boolean
    Dim img As Long, s As String
    Dim encoderCLSID As CLSID, encoderParams As EncoderParameters
    Dim WmfHeader As wmfPlaceableFileHeader

    If p Is Nothing Then Exit Function
    If GdiToken = 0 Then InitGDIPlus: bCloseGdiToken = True

    ' Determine pictyre type
    Select Case p.Type
        Case vbPicTypeBitmap
'             GdipCreateBitmapFromHBITMAP p.Handle, ByVal 0&, img
             GdipCreateBitmapFromHBITMAP p.Handle, p.hPal, img
        Case vbPicTypeMetafile
             FillInWmfHeader WmfHeader, p.Width, p.Height
             GdipCreateMetafileFromWmf p.Handle, False, WmfHeader, img
        Case vbPicTypeEMetafile
             GdipCreateMetafileFromEmf p.Handle, False, img
        Case vbPicTypeIcon
             ' Does not return a valid Image object
             GdipCreateBitmapFromHICON p.Handle, img
    End Select
    ' Load the Image
    If img Then
        ' Get the CLSID of the JPG encoder
        GetEncoderClsid GetFileType(FileType), encoderCLSID
        If FileType = enGdiFileType.JPG Then
            ' Save as a JPEG/GIF file. This format requires encoder parameters.
            ' Setup the encoder paramters
            With encoderParams
                With .Parameter
                    .NumberOfValues = 1     ' Should be one
                    .Type = EncoderParameterValueTypeLong
                    ' Set the GUI to EncoderQuality
                    GetParameterGUID EncoderQuality, .GUID
                    .Value = VarPtr(Quality)   ' Remember: The value expects only pointers!
                End With
                .Count = 1     ' Only one element in this Parameter array
            End With
        End If
        SavePictureToFile = (GdipSaveImageToFile(img, StrPtr(sFile), encoderCLSID, encoderParams) = GpStatus.OK)
        GdipDisposeImage img
    End If

    If bCloseGdiToken Then FreeGDIPlus: bCloseGdiToken = False
End Function

Public Function SavePictureToBytes(p As IPicture, m() As Byte, Optional FileType As enGdiFileType = JPG, Optional Quality As Long = 90) As Boolean
    Dim bCloseGdiToken As Boolean, Stream As IUnknown, hMem As Long, Size As Long
    Dim img As Long, s As String, b As Boolean
    Dim encoderCLSID As CLSID, encoderParams As EncoderParameters
    Dim WmfHeader As wmfPlaceableFileHeader
    Dim hGlobal As Long

    Erase m
    If GdiToken = 0 Then InitGDIPlus: bCloseGdiToken = True

    ' Determine pictyre type
    Select Case p.Type
        Case vbPicTypeBitmap
             GdipCreateBitmapFromHBITMAP p.Handle, ByVal 0&, img
        Case vbPicTypeMetafile
             FillInWmfHeader WmfHeader, p.Width, p.Height
             GdipCreateMetafileFromWmf p.Handle, False, WmfHeader, img
        Case vbPicTypeEMetafile
             GdipCreateMetafileFromEmf p.Handle, False, img
        Case vbPicTypeIcon
             ' Does not return a valid Image object
             GdipCreateBitmapFromHICON p.Handle, img
    End Select

    ' Load the Image
    If img Then
        ' Get the CLSID of the JPG encoder
        GetEncoderClsid GetFileType(FileType), encoderCLSID

        ' Save as a JPEG/GIF file. This format requires encoder parameters.
        ' Setup the encoder paramters
        If FileType = enGdiFileType.JPG Then
            With encoderParams
                With .Parameter
                    .NumberOfValues = 1     ' Should be one
                    .Type = EncoderParameterValueTypeLong
                    ' Set the GUI to EncoderQuality
                    GetParameterGUID EncoderQuality, .GUID
                    .Value = VarPtr(Quality)   ' Remember: The value expects only pointers!
                End With
                .Count = 1     ' Only one element in this Parameter array
            End With
        End If

        CreateStreamOnHGlobal ByVal 0&, 1, Stream
        b = (GdipSaveImageToStream(img, Stream, encoderCLSID, encoderParams) = GpStatus.OK)
        GdipDisposeImage img

        If b Then
            If GetHGlobalFromStream(Stream, hGlobal) = 0 Then
                ' Get the memory size
                Size = GlobalSize(hGlobal)
                ' Get a pointer to the memory
                hMem = GlobalLock(hGlobal)
                If hMem <> 0 And Size <> 0 Then
                    'Create the buffer and copy the data
                    ReDim m(0 To Size - 1)
                    CopyMemory m(0), ByVal GlobalLock(hMem), Size
                    SavePictureToBytes = True
                End If
                GlobalUnlock hMem
                GlobalFree hMem
            End If
        End If
        Set Stream = Nothing
    End If

    If bCloseGdiToken Then FreeGDIPlus: bCloseGdiToken = False
End Function

Private Function GetFileType(iType As enGdiFileType)
    Select Case iType
        Case enGdiFileType.BMP
            GetFileType = "bmp"
        Case enGdiFileType.GIF
            GetFileType = "gif"
        Case enGdiFileType.PNG
            GetFileType = "png"
        Case Else
            GetFileType = "jpeg"
    End Select
End Function

